# SPDX-FileCopyrightText: 2023-present Massimiliano Pippi <mpippi@gmail.com>
#
# SPDX-License-Identifier: MIT
import warnings

# banks depends on modules producing loads of deprecation warnings, let's just ignore them,
# nothing we can do anyways
warnings.simplefilter("ignore", category=DeprecationWarning)



import platform

import pytest

from banks import Prompt

from .conftest import anthropic_api_key_set, openai_api_key_set


def get_laptop_info():
    """Get information about the user laptop.

    For example, it returns the operating system and version, along with hardware and network specs."""
    return str(platform.uname())


@pytest.mark.e2e
@openai_api_key_set
def test_function_call_openai():
    p = Prompt("""
    {% set response %}
    {% completion model="gpt-3.5-turbo-0125" %}
        {% chat role="user" %}{{ query }}{% endchat %}
        {{ get_laptop_info | tool }}
    {% endcompletion %}
    {% endset %}

    {# the variable 'response' contains the result #}

    {{ response }}
    """)

    res = p.text({"query": "Can you guess the name of my laptop?", "get_laptop_info": get_laptop_info})
    assert res


@pytest.mark.e2e
@anthropic_api_key_set
def test_function_call_anthropic():
    p = Prompt("""
    {% set response %}
    {% completion model="claude-3-5-sonnet-20240620" %}
        {% chat role="user" %}{{ query }}{% endchat %}
        {{ get_laptop_info | tool }}
    {% endcompletion %}
    {% endset %}

    {# the variable 'response' contains the result #}

    {{ response }}
    """)

    res = p.text({"query": "Can you guess the name of my laptop? Use tools.", "get_laptop_info": get_laptop_info})
    assert res



import pytest
from jinja2 import TemplateSyntaxError

from banks import Prompt
from banks.extensions.chat import _ContentBlockParser
from banks.types import CacheControl, ContentBlock, ContentBlockType


def test_wrong_tag():
    with pytest.raises(TemplateSyntaxError):
        Prompt("{% chat %}{% endchat %}")


def test_wrong_tag_params():
    with pytest.raises(TemplateSyntaxError):
        Prompt('{% chat foo="bar" %}{% endchat %}')


def test_wrong_role_type():
    with pytest.raises(TemplateSyntaxError):
        Prompt('{% chat role="does not exist" %}{% endchat %}')


def test_content_block_parser_init():
    p = _ContentBlockParser()
    assert p._parse_block_content is False
    assert p._content_blocks == []


def test_content_block_parser_single_with_cache_control():
    p = _ContentBlockParser()
    p.feed(
        '<content_block>{"type":"text","cache_control":{"type":"ephemeral"},"text":"foo","source":null}</content_block>'
    )
    assert p.content == [
        ContentBlock(type=ContentBlockType.text, cache_control=CacheControl(type="ephemeral"), text="foo", source=None)
    ]


def test_content_block_parser_single_no_cache_control():
    p = _ContentBlockParser()
    p.feed('<content_block>{"type":"text","cache_control":null,"text":"foo","source":null}</content_block>')
    assert p.content == "foo"


def test_content_block_parser_multiple():
    p = _ContentBlockParser()
    p.feed(
        '<content_block>{"type":"text","cache_control":null,"text":"foo","source":null}</content_block>'
        '<content_block>{"type":"text","cache_control":null,"text":"bar","source":null}</content_block>'
    )
    assert p.content == [
        ContentBlock(type=ContentBlockType.text, cache_control=None, text="foo", source=None),
        ContentBlock(type=ContentBlockType.text, cache_control=None, text="bar", source=None),
    ]


def test_content_block_parser_other_tags():
    p = _ContentBlockParser()
    p.feed("<some_tag>FOO</some_tag>")
    assert p.content == "FOO"



# SPDX-FileCopyrightText: 2023-present Massimiliano Pippi <mpippi@gmail.com>
#
# SPDX-License-Identifier: MIT
"""
File-based prompt registry implementation that stores all prompts in a single JSON file.

This module provides functionality to store and retrieve prompts using a single JSON file
as the storage backend. The file contains an index of all prompts with their associated
metadata and content.
"""

from pathlib import Path

from pydantic import BaseModel

from banks.errors import InvalidPromptError, PromptNotFoundError
from banks.prompt import Prompt, PromptModel


class PromptRegistryIndex(BaseModel):
    """
    Model representing the registry index containing all prompts.

    Stores a list of PromptModel objects that represent all prompts in the registry.
    """

    prompts: list[PromptModel] = []


class FilePromptRegistry:
    """A prompt registry storing all prompt data in a single JSON file."""

    def __init__(self, registry_index: Path) -> None:
        """
        Initialize the file prompt registry.

        Args:
            registry_index: Path to the JSON file that will store the prompts

        Note:
            Creates parent directories if they don't exist.
        """
        self._index_fpath: Path = registry_index
        self._index: PromptRegistryIndex = PromptRegistryIndex(prompts=[])
        try:
            self._index = PromptRegistryIndex.model_validate_json(self._index_fpath.read_text())
        except FileNotFoundError:
            # init the user data folder
            self._index_fpath.parent.mkdir(parents=True, exist_ok=True)

    def get(self, *, name: str, version: str | None = None) -> Prompt:
        """
        Retrieve a prompt by name and version.

        Args:
            name: Name of the prompt to retrieve
            version: Version of the prompt (optional)

        Returns:
            The requested Prompt object

        Raises:
            PromptNotFoundError: If the requested prompt doesn't exist
        """
        _, model = self._get_prompt_model(name, version)
        return Prompt(**model.model_dump())

    def set(self, *, prompt: Prompt, overwrite: bool = False) -> None:
        """
        Store a prompt in the registry.

        Args:
            prompt: The Prompt object to store
            overwrite: Whether to overwrite an existing prompt

        Raises:
            InvalidPromptError: If prompt exists and overwrite=False
        """
        try:
            idx, p_model = self._get_prompt_model(prompt.name, prompt.version)
            if overwrite:
                self._index.prompts[idx] = PromptModel.from_prompt(prompt)
                self._save()
            else:
                msg = f"Prompt with name '{prompt.name}' already exists. Use overwrite=True to overwrite"
                raise InvalidPromptError(msg)
        except PromptNotFoundError:
            p_model = PromptModel.from_prompt(prompt)
            self._index.prompts.append(p_model)
            self._save()

    def _save(self) -> None:
        """
        Save the prompt index to the JSON file.

        Writes the current state of the registry to disk.
        """
        with open(self._index_fpath, "w", encoding="locale") as f:
            f.write(self._index.model_dump_json())

    def _get_prompt_model(self, name: str | None, version: str | None) -> tuple[int, PromptModel]:
        """
        Find a prompt model in the index by name and version.

        Args:
            name: Name of the prompt
            version: Version of the prompt

        Returns:
            Tuple of (index position, PromptModel)

        Raises:
            PromptNotFoundError: If the prompt doesn't exist in the index
        """
        for i, model in enumerate(self._index.prompts):
            if model.name == name and model.version == version:
                return i, model

        msg = f"cannot find prompt with name '{name}' and version '{version}'"
        raise PromptNotFoundError(msg)



# SPDX-FileCopyrightText: 2023-present Massimiliano Pippi <mpippi@gmail.com>
#
# SPDX-License-Identifier: MIT
from html.parser import HTMLParser

from jinja2 import TemplateSyntaxError, nodes
from jinja2.ext import Extension

from banks.types import ChatMessage, ChatMessageContent, ContentBlock, ContentBlockType

SUPPORTED_TYPES = ("system", "user")


class _ContentBlockParser(HTMLParser):
    """A parser used to extract text surrounded by `<content_block_txt>` and `</content_block_txt>` tags."""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._parse_block_content = False
        self._content_blocks: list[ContentBlock] = []

    @property
    def content(self) -> ChatMessageContent:
        """Returns ChatMessageContent data that can be directly assigned to ChatMessage.content.

        If only one block is present, this block is of type text and has no cache control set, we just
        return it as plain text for simplicity.
        """
        if len(self._content_blocks) == 1:
            block = self._content_blocks[0]
            if block.type == "text" and block.cache_control is None:
                return block.text or ""

        return self._content_blocks

    def handle_starttag(self, tag, attrs):  # noqa
        if tag == "content_block":
            self._parse_block_content = True

    def handle_endtag(self, tag):
        if tag == "content_block":
            self._parse_block_content = False

    def handle_data(self, data):
        if self._parse_block_content:
            self._content_blocks.append(ContentBlock.model_validate_json(data))
        else:
            self._content_blocks.append(ContentBlock(type=ContentBlockType.text, text=data))


class ChatExtension(Extension):
    """
    `chat` can be used to render prompt text as structured ChatMessage objects.

    Example:
        ```
        {% chat role="system" %}
        You are a helpful assistant.
        {% endchat %}
        ```
    """

    # a set of names that trigger the extension.
    tags = {"chat"}  # noqa

    def parse(self, parser):
        # We get the line number of the first token for error reporting
        lineno = next(parser.stream).lineno

        # Gather tokens up to the next block_end ('%}')
        gathered = []
        while parser.stream.current.type != "block_end":
            gathered.append(next(parser.stream))

        # If all has gone well, we will have one triplet of tokens:
        #   (type='name, value='role'),
        #   (type='assign', value='='),
        #   (type='string', value='user'),
        # Anything else is a parse error
        error_msg = f"Invalid syntax for chat attribute, got '{gathered}', expected role=\"value\""
        try:
            attr_name, attr_assign, attr_value = gathered  # pylint: disable=unbalanced-tuple-unpacking
        except ValueError:
            raise TemplateSyntaxError(error_msg, lineno) from None

        # Validate tag attributes
        if attr_name.value != "role" or attr_assign.value != "=":
            raise TemplateSyntaxError(error_msg, lineno)

        if attr_value.value not in SUPPORTED_TYPES:
            types = ",".join(SUPPORTED_TYPES)
            msg = f"Unknown role type '{attr_value}', use one of ({types})"
            raise TemplateSyntaxError(msg, lineno)

        # Pass the role name to the CallBlock node
        args: list[nodes.Expr] = [nodes.Const(attr_value.value)]

        # Message body
        body = parser.parse_statements(("name:endchat",), drop_needle=True)

        # Build messages list
        return nodes.CallBlock(self.call_method("_store_chat_messages", args), [], [], body).set_lineno(lineno)

    def _store_chat_messages(self, role, caller):
        """
        Helper callback.
        """
        parser = _ContentBlockParser()
        parser.feed(caller())
        cm = ChatMessage(role=role, content=parser.content)
        return cm.model_dump_json(exclude_none=True) + "\n"



from banks.filters import tool
from banks.types import Tool


def test_tool():
    def my_tool_function(myparam: str):
        """Description of the tool.

        Args:
            myparam (str): description of the parameter

        """
        pass

    tool_dump = tool(my_tool_function)
    t = Tool.model_validate_json(tool_dump)
    assert t.model_dump() == {
        "type": "function",
        "function": {
            "name": "my_tool_function",
            "description": "Description of the tool.\n\nArgs:\n    myparam (str): description of the parameter",
            "parameters": {
                "type": "object",
                "properties": {"myparam": {"type": "string", "description": "description of the parameter"}},
                "required": ["myparam"],
            },
        },
        "import_path": "tests.test_tool.test_tool.<locals>.my_tool_function",
    }


def test_tool_with_defaults():
    def my_tool_function(myparam: str = ""):
        """Description of the tool.

        Args:
            myparam (str): description of the parameter

        """
        pass

    tool_dump = tool(my_tool_function)
    t = Tool.model_validate_json(tool_dump)
    assert t.model_dump() == {
        "type": "function",
        "function": {
            "name": "my_tool_function",
            "description": "Description of the tool.\n\nArgs:\n    myparam (str): description of the parameter",
            "parameters": {
                "type": "object",
                "properties": {"myparam": {"type": "string", "description": "description of the parameter"}},
                "required": [],
            },
        },
        "import_path": "tests.test_tool.test_tool_with_defaults.<locals>.my_tool_function",
    }



import pytest

from banks.errors import InvalidPromptError, PromptNotFoundError
from banks.prompt import Prompt, PromptModel
from banks.registries.file import FilePromptRegistry, PromptRegistryIndex


@pytest.fixture
def populated_registry(tmp_path):
    prompt_models = [PromptModel(name="name", version="version", text="prompt")]
    idx = PromptRegistryIndex(prompts=prompt_models)
    with open(tmp_path / "index.json", "w") as f:
        f.write(idx.model_dump_json())
    return FilePromptRegistry(tmp_path / "index.json")


def test_init_from_scratch(tmp_path):
    index_file = tmp_path / "test" / "index.json"
    r = FilePromptRegistry(index_file)
    assert r._index_fpath == index_file
    assert index_file.parent.exists()


def test_init_from_existing_dir(tmp_path):
    r = FilePromptRegistry(tmp_path / "index.json")
    r._save()
    assert r._index_fpath.parent.exists()


def test_init_from_existing_index(populated_registry):
    populated_registry.get(name="name", version="version")


def test_get_not_found(populated_registry):
    with pytest.raises(PromptNotFoundError):
        populated_registry.get(name="name", version="nonexisting_version")


def test_set_existing_no_overwrite(populated_registry):
    populated_registry.set(prompt=Prompt("a new prompt!", name="foo"))
    with pytest.raises(InvalidPromptError):
        populated_registry.set(prompt=Prompt("a new prompt!", name="foo"))


def test_set_existing_overwrite(populated_registry):
    new_prompt = "a new prompt!"
    populated_registry.set(prompt=Prompt(new_prompt, name="name", version="version"), overwrite=True)
    assert populated_registry.get(name="name", version="version").raw == new_prompt


def test_set_new(populated_registry):
    new_prompt = "a new prompt!"
    populated_registry.set(prompt=Prompt(new_prompt, name="name", version="version2"))
    assert populated_registry.get(name="name", version="version").raw == "prompt"
    assert populated_registry.get(name="name", version="version2").raw == new_prompt



# SPDX-FileCopyrightText: 2023-present Massimiliano Pippi <mpippi@gmail.com>
#
# SPDX-License-Identifier: MIT
import importlib
import json
from typing import cast

from jinja2 import TemplateSyntaxError, nodes
from jinja2.ext import Extension
from litellm import acompletion, completion
from litellm.types.utils import Choices, ModelResponse
from pydantic import ValidationError

from banks.errors import InvalidPromptError, LLMError
from banks.types import ChatMessage, Tool

SUPPORTED_KWARGS = ("model",)


class CompletionExtension(Extension):
    """
    `completion` can be used to send to the LLM the content of the block in form of messages.

    The rendered value of the block can be used as is but it's usually more useful to
    assign it to a variable and access it from another section of the prompt.

    Example:
        ```
        {% set response %}
        {% completion model="gpt-3.5-turbo-0125" %}
        {% chat role="user" %}You are a helpful assistant{% endchat %}
        {% endcompletion %}
        {% endset %}

        {# output the response content #}
        {{ response }}
        ```
    """

    # a set of names that trigger the extension.
    tags = {"completion"}  # noqa

    def parse(self, parser):
        # We get the line number of the first token for error reporting
        lineno = next(parser.stream).lineno

        # Gather tokens up to the next block_end ('%}')
        gathered = []
        while parser.stream.current.type != "block_end":
            gathered.append(next(parser.stream))

        # If all has gone well, we will have one triplet of tokens:
        #   (type='name', value='model'),
        #   (type='assign', value='='),
        #   (type='string', value='gpt-3.5-turbo-0125'),
        # Anything else is a parse error
        error_msg = f"Invalid syntax for completion: {gathered}"
        try:
            attr_name, attr_assign, attr_value = gathered  # pylint: disable=unbalanced-tuple-unpacking
        except ValueError:
            raise TemplateSyntaxError(error_msg, lineno) from None

        # Validate tag attributes
        if attr_name.value not in SUPPORTED_KWARGS or attr_assign.value != "=":
            raise TemplateSyntaxError(error_msg, lineno)

        # Pass the role name to the CallBlock node
        args: list[nodes.Expr] = [nodes.Const(attr_value.value)]

        # Message body
        body = parser.parse_statements(("name:endcompletion",), drop_needle=True)

        # Call LLM
        if parser.environment.is_async:
            return nodes.CallBlock(self.call_method("_do_completion_async", args), [], [], body).set_lineno(lineno)
        return nodes.CallBlock(self.call_method("_do_completion", args), [], [], body).set_lineno(lineno)

    def _get_tool_callable(self, tools, tool_call):
        for tool in tools:
            if tool.function.name == tool_call.function.name:
                module_name, func_name = tool.import_path.rsplit(".", maxsplit=1)
                module = importlib.import_module(module_name)
                return getattr(module, func_name)
        msg = f"Function {tool_call.function.name} not found in available tools"
        raise ValueError(msg)

    def _do_completion(self, model_name, caller):
        """
        Helper callback.
        """
        messages, tools = self._body_to_messages(caller())
        message_dicts = [m.model_dump() for m in messages]
        tool_dicts = [t.model_dump() for t in tools] or None

        response = cast(ModelResponse, completion(model=model_name, messages=message_dicts, tools=tool_dicts))
        choices = cast(list[Choices], response.choices)
        tool_calls = choices[0].message.tool_calls
        if not tool_calls:
            return choices[0].message.content

        message_dicts.append(choices[0].message.model_dump())
        for tool_call in tool_calls:
            if not tool_call.function.name:
                msg = "Malformed response: function name is empty"
                raise LLMError(msg)

            func = self._get_tool_callable(tools, tool_call)

            function_args = json.loads(tool_call.function.arguments)
            function_response = func(**function_args)
            message_dicts.append(
                ChatMessage(
                    tool_call_id=tool_call.id, role="tool", name=tool_call.function.name, content=function_response
                ).model_dump()
            )

        response = cast(ModelResponse, completion(model=model_name, messages=message_dicts, tools=tool_dicts))
        choices = cast(list[Choices], response.choices)
        return choices[0].message.content

    async def _do_completion_async(self, model_name, caller):
        """
        Helper callback.
        """
        messages, tools = self._body_to_messages(caller())
        message_dicts = [m.model_dump() for m in messages]
        tool_dicts = [t.model_dump() for t in tools] or None

        response = cast(ModelResponse, await acompletion(model=model_name, messages=message_dicts, tools=tool_dicts))
        choices = cast(list[Choices], response.choices)
        tool_calls = choices[0].message.tool_calls or []
        if not tool_calls:
            return choices[0].message.content

        message_dicts.append(choices[0].message.model_dump())
        for tool_call in tool_calls:
            if not tool_call.function.name:
                msg = "Function name is empty"
                raise LLMError(msg)

            func = self._get_tool_callable(tools, tool_call)

            message_dicts.append(
                ChatMessage(
                    tool_call_id=tool_call.id,
                    role="tool",
                    name=tool_call.function.name,
                    content=func(**json.loads(tool_call.function.arguments)),
                ).model_dump()
            )

        response = cast(ModelResponse, await acompletion(model=model_name, messages=message_dicts, tools=tool_dicts))
        choices = cast(list[Choices], response.choices)
        return choices[0].message.content

    def _body_to_messages(self, body: str) -> tuple[list[ChatMessage], list[Tool]]:
        """Converts each line in the body of a block into a chat message."""
        body = body.strip()
        messages = []
        tools = []
        for line in body.split("\n"):
            try:
                # Try to parse a chat message
                messages.append(ChatMessage.model_validate_json(line))
            except ValidationError:  # pylint: disable=R0801
                try:
                    # If not a chat message, try to parse a tool
                    tools.append(Tool.model_validate_json(line))
                except ValidationError:
                    # Give up
                    pass

        if not messages:
            msg = "Completion must contain at least one chat message"
            raise InvalidPromptError(msg)

        return (messages, tools)



from banks.filters.cache_control import cache_control


def test_cache_control():
    res = cache_control("foo", "ephemeral")
    res = res.replace("<content_block>", "")
    res = res.replace("</content_block>", "")
    assert res == '{"type":"text","cache_control":{"type":"ephemeral"},"text":"foo","image_url":null}'



import pytest

from banks.cache import DefaultCache


@pytest.fixture
def cache():
    return DefaultCache()


@pytest.mark.parametrize(
    "context_value",
    [{}, {"foo": "bar"}],
)
def test_default_cache_set(context_value, cache):
    cache.set(context_value, "My prompt")
    assert len(cache._cache) == 1
    assert next(iter(cache._cache.values())) == "My prompt"


def test_default_cache_get(cache):
    cache.set({"foo": "bar"}, "My prompt")
    assert cache.get({"foo": "bar"}) == "My prompt"
    assert cache.get({"bar"}) is None


def test_default_cache_clear(cache):
    cache.set({"foo": "bar"}, "My prompt")
    cache.clear()
    assert not len(cache._cache)



# SPDX-FileCopyrightText: 2023-present Massimiliano Pippi <mpippi@gmail.com>
#
# SPDX-License-Identifier: MIT
from typing import Callable

from banks.types import Tool


def tool(function: Callable) -> str:
    """Inspect a Python callable and generates a JSON-schema ready for LLM function calling.

    Important:
        This filter only works when used within a `{% completion %}` block.
    """
    t = Tool.from_callable(function)
    return t.model_dump_json() + "\n"



# SPDX-FileCopyrightText: 2023-present Massimiliano Pippi <mpippi@gmail.com>
#
# SPDX-License-Identifier: MIT
class MissingDependencyError(Exception):
    """Some optional dependencies are missing."""


class AsyncError(Exception):
    """An error related to asyncio support."""


class CanaryWordError(Exception):
    """The canary word has leaked."""


class PromptNotFoundError(Exception):
    """The prompt was now found in the registry."""


class InvalidPromptError(Exception):
    """The prompt is not valid."""


class LLMError(Exception):
    """The LLM had problems."""



# SPDX-FileCopyrightText: 2023-present Massimiliano Pippi <mpippi@gmail.com>
#
# SPDX-License-Identifier: MIT
from pathlib import Path
from urllib.parse import urlparse

from banks.types import ContentBlock, ImageUrl


def _is_url(string: str) -> bool:
    result = urlparse(string)
    return all([result.scheme, result.netloc])


def image(value: str) -> str:
    """Wrap the filtered value into a ContentBlock of type image.

    The resulting ChatMessage will have the field `content` populated with a list of ContentBlock objects.

    Example:
        ```jinja
        Describe what you see

        {{ "path/to/image/file" | image }}
        ```

    Important:
        this filter marks the content to cache by surrounding it with `<content_block>` and
        `</content_block>`, so it's only useful when used within a `{% chat %}` block.
    """
    if _is_url(value):
        image_url = ImageUrl(url=value)
    else:
        image_url = ImageUrl.from_path(Path(value))

    block = ContentBlock.model_validate({"type": "image_url", "image_url": image_url})
    return f"<content_block>{block.model_dump_json()}</content_block>"



import os

import pytest

openai_api_key_set = pytest.mark.skipif(not os.environ.get("OPENAI_API_KEY"), reason="OPENAI_API_KEY not set")
anthropic_api_key_set = pytest.mark.skipif(not os.environ.get("ANTHROPIC_API_KEY"), reason="ANTHROPIC_API_KEY not set")



# SPDX-FileCopyrightText: 2023-present Massimiliano Pippi <mpippi@gmail.com>
#
# SPDX-License-Identifier: MIT
"""
Directory-based prompt registry implementation that stores prompts as files.
"""

import time
from pathlib import Path

try:
    from typing import Self
except ImportError:  # pragma: no cover
    from typing_extensions import Self
from pydantic import BaseModel, Field

from banks import Prompt
from banks.errors import InvalidPromptError, PromptNotFoundError
from banks.prompt import DEFAULT_VERSION, PromptModel

# Constants
DEFAULT_INDEX_NAME = "index.json"


class PromptFile(PromptModel):
    """Model representing a prompt file stored on disk."""

    path: Path = Field(exclude=True)

    @classmethod
    def from_prompt_path(cls: type[Self], prompt: Prompt, path: Path) -> Self:
        """
        Create a PromptFile instance from a Prompt object and save it to disk.

        Args:
            prompt: The Prompt object to save
            path: Directory path where the prompt file should be stored

        Returns:
            A new PromptFile instance
        """
        prompt_file = path / f"{prompt.name}.{prompt.version}.jinja"
        prompt_file.write_text(prompt.raw)
        return cls(
            text=prompt.raw, name=prompt.name, version=prompt.version, metadata=prompt.metadata, path=prompt_file
        )


class PromptFileIndex(BaseModel):
    """Index tracking all prompt files in the directory."""

    files: list[PromptFile] = Field(default=[])


class DirectoryPromptRegistry:
    """Registry that stores prompts as files in a directory structure."""

    def __init__(self, directory_path: Path, *, force_reindex: bool = False):
        """
        Initialize the directory prompt registry.

        Args:
            directory_path: Path to directory where prompts will be stored
            force_reindex: Whether to force rebuilding the index from disk

        Raises:
            ValueError: If directory_path is not a directory
        """
        if not directory_path.is_dir():
            msg = "{directory_path} must be a directory."
            raise ValueError(msg)

        self._path = directory_path
        self._index_path = self._path / DEFAULT_INDEX_NAME
        if not self._index_path.exists() or force_reindex:
            self._scan()
        else:
            self._load()

    @property
    def path(self) -> Path:
        """Get the directory path where prompts are stored."""
        return self._path

    def get(self, *, name: str, version: str | None = None) -> Prompt:
        """
        Retrieve a prompt by name and version.

        Args:
            name: Name of the prompt to retrieve
            version: Version of the prompt (optional)

        Returns:
            The requested Prompt object

        Raises:
            PromptNotFoundError: If prompt doesn't exist
        """
        version = version or DEFAULT_VERSION
        for pf in self._index.files:
            if pf.name == name and pf.version == version and pf.path.exists():
                return Prompt(**pf.model_dump())
        raise PromptNotFoundError

    def set(self, *, prompt: Prompt, overwrite: bool = False):
        """
        Store a prompt in the registry.

        Args:
            prompt: The Prompt object to store
            overwrite: Whether to overwrite existing prompt

        Raises:
            InvalidPromptError: If prompt exists and overwrite=False
        """
        try:
            version = prompt.version or DEFAULT_VERSION
            idx, pf = self._get_prompt_file(name=prompt.name, version=version)
            if overwrite:
                prompt.metadata["created_at"] = time.ctime()
                self._index.files[idx] = PromptFile.from_prompt_path(prompt, self._path)
                self._save()
            else:  # pylint: disable=duplicate-code
                msg = f"Prompt with name '{prompt.name}' already exists. Use overwrite=True to overwrite"
                raise InvalidPromptError(msg)
        except PromptNotFoundError:
            prompt.metadata["created_at"] = time.ctime()
            pf = PromptFile.from_prompt_path(prompt, self._path)
            self._index.files.append(pf)
            self._save()

    def _load(self):
        """Load the prompt index from disk."""
        self._index = PromptFileIndex.model_validate_json(self._index_path.read_text())

    def _save(self):
        """Save the prompt index to disk."""
        self._index_path.write_text(self._index.model_dump_json())

    def _scan(self):
        """Scan directory for prompt files and build the index."""
        self._index: PromptFileIndex = PromptFileIndex()
        for path in self._path.glob("*.jinja*"):
            name, version = path.stem.rsplit(".", 1) if "." in path.stem else (path.stem, DEFAULT_VERSION)
            with path.open("r") as f:
                pf = PromptFile(text=f.read(), name=name, version=version, path=path, metadata={})
                self._index.files.append(pf)
        self._index_path.write_text(self._index.model_dump_json())

    def _get_prompt_file(self, *, name: str | None, version: str) -> tuple[int, PromptFile]:
        """
        Find a prompt file in the index.

        Args:
            name: Name of the prompt
            version: Version of the prompt

        Returns:
            Tuple of (index position, PromptFile)

        Raises:
            PromptNotFoundError: If prompt doesn't exist in index
        """
        for i, pf in enumerate(self._index.files):
            if pf.name == name and pf.version == version:
                return i, pf

        msg = f"cannot find prompt with name '{name}' and version '{version}'"
        raise PromptNotFoundError(msg)



# SPDX-FileCopyrightText: 2023-present Massimiliano Pippi <mpippi@gmail.com>
#
# SPDX-License-Identifier: MIT
from banks.errors import MissingDependencyError

try:
    from simplemma import text_lemmatizer  # type: ignore

    SIMPLEMMA_AVAIL = True
except ImportError:  # pragma: no cover
    SIMPLEMMA_AVAIL = False


def lemmatize(text: str) -> str:
    """
    Compute and return the lemmatization of the input. Language is hardcoded
    to English.

    Example:
        ```jinja
        {{"The dog is running" | lemmatize}}
        "the dog be run"
        ```

    Note:
        Simplemma must be manually installed to use this filter
    """
    if not SIMPLEMMA_AVAIL:
        err_msg = "simplemma is not available, please install it with 'pip install simplemma'"
        raise MissingDependencyError(err_msg)

    words = text_lemmatizer(text, "en")
    return " ".join(words)



# SPDX-FileCopyrightText: 2023-present Massimiliano Pippi <mpippi@gmail.com>
#
# SPDX-License-Identifier: MIT
from jinja2 import Environment, select_autoescape

from .config import config
from .filters import cache_control, image, lemmatize, tool


def _add_extensions(_env):
    """
    We lazily add extensions so that we can use the env in the extensions themselves if needed.

    For example, we use banks to manage the system prompt in `GenerateExtension`
    """
    from .extensions.chat import ChatExtension  # pylint: disable=import-outside-toplevel
    from .extensions.completion import CompletionExtension  # pylint: disable=import-outside-toplevel

    _env.add_extension(ChatExtension)
    _env.add_extension(CompletionExtension)


# Init the Jinja env
env = Environment(
    autoescape=select_autoescape(
        enabled_extensions=("html", "xml"),
        default_for_string=False,
    ),
    trim_blocks=True,
    lstrip_blocks=True,
    enable_async=bool(config.ASYNC_ENABLED),
)


# Setup custom filters and defaults
env.filters["cache_control"] = cache_control
env.filters["image"] = image
env.filters["lemmatize"] = lemmatize
env.filters["tool"] = tool

_add_extensions(env)



# SPDX-FileCopyrightText: 2023-present Massimiliano Pippi <mpippi@gmail.com>
#
# SPDX-License-Identifier: MIT
import base64
from enum import Enum
from inspect import Parameter, getdoc, signature
from pathlib import Path
from typing import Callable

from pydantic import BaseModel
from typing_extensions import Self

from .utils import parse_params_from_docstring, python_type_to_jsonschema

# pylint: disable=invalid-name


class ContentBlockType(str, Enum):
    text = "text"
    image_url = "image_url"


class CacheControl(BaseModel):
    type: str = "ephemeral"


class ImageUrl(BaseModel):
    url: str

    @classmethod
    def from_base64(cls, media_type: str, base64_str: str) -> Self:
        return cls(url=f"data:{media_type};base64,{base64_str}")

    @classmethod
    def from_path(cls, file_path: Path) -> Self:
        with open(file_path, "rb") as image_file:
            return cls.from_base64("image/jpeg", base64.b64encode(image_file.read()).decode("utf-8"))


class ContentBlock(BaseModel):
    type: ContentBlockType
    cache_control: CacheControl | None = None
    text: str | None = None
    image_url: ImageUrl | None = None


ChatMessageContent = list[ContentBlock] | str


class ChatMessage(BaseModel):
    role: str
    content: str | ChatMessageContent
    tool_call_id: str | None = None
    name: str | None = None


class FunctionParameter(BaseModel):
    type: str
    description: str


class FunctionParameters(BaseModel):
    type: str = "object"
    properties: dict[str, FunctionParameter]
    required: list[str]


class Function(BaseModel):
    name: str
    description: str
    parameters: FunctionParameters


class Tool(BaseModel):
    """A model representing a Tool to be used in function calling.

    This model should dump the following:
    ```
    {
        "type": "function",
        "function": {
            "name": "get_current_weather",
            "description": "Get the current weather in a given location",
            "parameters": {
                "type": "object",
                "properties": {
                    "location": {
                        "type": "string",
                        "description": "The city and state, e.g. San Francisco, CA",
                    },
                    "unit": {"type": "string", "enum": ["celsius", "fahrenheit"]},
                },
                "required": ["location"],
            },
        },
        "import_path": "module.get_current_weather",
    }
    ```
    """

    type: str = "function"
    function: Function
    import_path: str

    @classmethod
    def from_callable(cls, func: Callable) -> Self:
        sig = signature(func)

        # Try getting params descriptions from docstrings
        param_docs = parse_params_from_docstring(func.__doc__ or "")

        # If the docstring is missing, use the qualname space-separated hopefully
        # the LLM will get some more context than just using the function name.
        description = getdoc(func) or " ".join(func.__qualname__.split("."))
        properties = {}
        required = []
        for name, param in sig.parameters.items():
            p = FunctionParameter(
                description=param_docs.get(name, {}).get("description", ""),
                type=python_type_to_jsonschema(param.annotation),
            )
            properties[name] = p
            if param.default == Parameter.empty:
                required.append(name)

        return cls(
            function=Function(
                name=func.__name__,
                description=description,
                parameters=FunctionParameters(properties=properties, required=required),
            ),
            import_path=f"{func.__module__}.{func.__qualname__}",
        )



# SPDX-FileCopyrightText: 2023-present Massimiliano Pippi <mpippi@gmail.com>
#
# SPDX-License-Identifier: MIT
import pickle
from typing import Optional, Protocol, runtime_checkable


@runtime_checkable
class RenderCache(Protocol):  # pragma: no cover
    """
    Interface for rendering cache.

    Prompts with the same template and same context are retrieved from the cache directly,
    instead of going through the rendering process.
    """

    def get(self, context: dict) -> Optional[str]: ...

    def set(self, context: dict, prompt: str) -> None: ...

    def clear(self) -> None: ...


class DefaultCache:
    """
    In-memory, default rendering cache.
    """

    def __init__(self) -> None:
        self._cache: dict[bytes, str] = {}

    def get(self, context: dict) -> Optional[str]:
        return self._cache.get(pickle.dumps(context, pickle.HIGHEST_PROTOCOL))

    def set(self, context: dict, prompt: str) -> None:
        self._cache[pickle.dumps(context, pickle.HIGHEST_PROTOCOL)] = prompt

    def clear(self) -> None:
        self._cache = {}



# This module exists for documentation purpose only


def chat(role: str):  # pylint: disable=W0613
    """
    Text inside `chat` tags will be rendered as JSON strings representing chat messages. Calling `Prompt.chat_messages`
    will return a list of `ChatMessage` instances.

    Example:
        ```jinja
        {% chat role="system" %}
        You are a helpful assistant.
        {% endchat %}

        {% chat role="user" %}
        Hello, how are you?
        {% endchat %}
        ```
    """


def completion(model_name: str):  # pylint: disable=W0613
    """
    `completion` can be used to send to the LLM the content of the block in form of messages.

    The rendered value of the block can be assigned to a variable and accessed from another section of the prompt.

    Example:
        ```jinja
        {% set response %}
        {% completion model="gpt-3.5-turbo-0125" %}
        {% chat role="user" %}You are a helpful assistant{% endchat %}
        {% endcompletion %}
        {% endset %}

        {# output the response content #}
        {{ response }}
        ```
    """



# SPDX-FileCopyrightText: 2023-present Massimiliano Pippi <mpippi@gmail.com>
#
# SPDX-License-Identifier: MIT
from .cache_control import cache_control
from .image import image
from .lemmatize import lemmatize
from .tool import tool

__all__ = (
    "cache_control",
    "image",
    "lemmatize",
    "tool",
)



import os
from pathlib import Path

import pytest

from banks.errors import InvalidPromptError, PromptNotFoundError
from banks.prompt import Prompt
from banks.registries.directory import DEFAULT_INDEX_NAME, DirectoryPromptRegistry, PromptFileIndex


@pytest.fixture
def registry(tmp_path: Path):
    d = tmp_path / "templates"
    d.mkdir()
    for fp in (Path(__file__).parent / "templates").iterdir():
        with open(d / fp.name, "w") as f:
            f.write(fp.read_text())
    return DirectoryPromptRegistry(d, force_reindex=True)


def test_init_from_scratch(registry: DirectoryPromptRegistry):
    p = registry.get(name="blog")
    assert p.raw.startswith("{# Zero-shot, this is already enough for most topics in english -#}")
    assert p.metadata == {}


def test_init_from_existing_index(tmp_path: Path):
    pi = PromptFileIndex()
    idx = tmp_path / DEFAULT_INDEX_NAME
    idx.write_text(pi.model_dump_json())
    r = DirectoryPromptRegistry(tmp_path)
    assert len(r._index.files) == 0


def test_init_from_existing_index_force(registry: DirectoryPromptRegistry):
    # change the directory structure
    f = registry.path / "blog.jinja"
    os.remove(f)
    # force recreation, the renamed file should be updated in the index
    r = DirectoryPromptRegistry(registry.path, force_reindex=True)
    with pytest.raises(PromptNotFoundError):
        r.get(name="blog")


def test_init_invalid_dir():
    with pytest.raises(ValueError):
        DirectoryPromptRegistry(Path("does/not/exists"))


def test_get_not_found(registry: DirectoryPromptRegistry):
    with pytest.raises(PromptNotFoundError):
        registry.get(name="FOO")


def test_set_existing_no_overwrite(registry: DirectoryPromptRegistry):
    new_prompt = Prompt("a new prompt!", name="blog")
    with pytest.raises(
        InvalidPromptError, match="Prompt with name 'blog' already exists. Use overwrite=True to overwrite"
    ):
        registry.set(prompt=new_prompt)


def test_set_existing_overwrite(registry: DirectoryPromptRegistry):
    new_prompt = Prompt("a new prompt!", name="blog")
    registry.set(prompt=new_prompt, overwrite=True)
    assert registry.get(name="blog").text() == "a new prompt!"
    assert "created_at" in registry.get(name="blog").metadata  # created_at is added when overwrite==True


def test_set_multiple_templates(registry: DirectoryPromptRegistry):
    new_prompt = Prompt("a very new prompt!", name="new", version="2")
    registry.set(prompt=new_prompt)
    old_prompt = Prompt("an old prompt!", name="old", version="1")
    registry.set(prompt=old_prompt)

    p = registry.get(name="old", version="1")
    assert p.raw == "an old prompt!"
    assert "created_at" in p.metadata

    p = registry.get(name="new", version="2")
    assert p.raw == "a very new prompt!"
    assert "created_at" in p.metadata


def test_update_meta(registry: DirectoryPromptRegistry):
    # test metadata for initial set
    new_prompt = Prompt("a very new prompt!", name="new", version="3", metadata={"accuracy": 91.2})
    registry.set(prompt=new_prompt)
    created_at = new_prompt.metadata["created_at"]
    assert registry.get(name="new", version="3").metadata == {"accuracy": 91.2, "created_at": created_at}

    # test metadata update for existing prompt
    p = registry.get(name="new", version="3")
    created_time = p.metadata["created_at"]
    p.metadata["accuracy"] = 94.3
    registry.set(prompt=p, overwrite=True)
    # reload prompt
    p = registry.get(name="new", version="3")
    assert p.metadata == {"accuracy": 94.3, "created_at": created_time}



# SPDX-FileCopyrightText: 2023-present Massimiliano Pippi <mpippi@gmail.com>
#
# SPDX-License-Identifier: MIT
import os
from pathlib import Path
from typing import Any

from platformdirs import user_data_path

from .utils import strtobool


class _BanksConfig:
    ASYNC_ENABLED: bool = False
    USER_DATA_PATH: Path = user_data_path("banks")

    def __init__(self, env_var_prefix: str = "BANKS_"):
        self._env_var_prefix = env_var_prefix

    def __getattribute__(self, name: str) -> Any:
        # Raise an attribute error if the name of the config is unknown
        original_value = super().__getattribute__(name)

        # Env var takes precedence
        prefix = super().__getattribute__("_env_var_prefix")
        read_value = os.environ.get(f"{prefix}{name}")
        if read_value is None:
            return original_value

        # Convert string from env var to the actual type
        t = super().__getattribute__("__annotations__")[name]
        if t is bool:
            return strtobool(read_value)

        return t(read_value)


config = _BanksConfig()
